package MPA;

import kqml.Handler;
import kqml.KqmlAgent;
import kqml.Message;
import kqml.Reply;
import kqml.Log;
import java.util.Vector;
import java.util.StringTokenizer;
import Lisp.List;

import transform.My_Util;
import transform.Reader;

/**
 * A class to provide KQML-based communication with other agents in SRI
 * MultiAgent Planning Architecture.
 *
 * @author Jim Shoop, ISX Corporation
 * @version $Revision: 0.1 $
 */
public class MPA_AgentImpl {
  
  ////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////

  // Debugging & Logging Switches
  private static final boolean	DEBUG			= true;
  private static final boolean	LOGGING			= true;
  private static final int	LOGGING_LEVEL		= Log.ALWAYS;

  // Agent Names
  private static final String   AGENT_DUMMY		= "";
  private static final String	AGENT_APAT		= AGENT_DUMMY + "APAT";
  private static final String	AGENT_PLAN_SERVER	= AGENT_DUMMY + "plan-server";
  private static final String	AGENT_PLAN_CELL_MGR	= AGENT_DUMMY + "meta-pcm";
  private static final String	AGENT_ACS		= AGENT_DUMMY + "ACS";

  // KQML Performatives
  private static final String	PERFORMATIVE_EVALUATE	= "evaluate";
  private static final String	PERFORMATIVE_ASK_ALL	= "ask-all";
  private static final String	PERFORMATIVE_TELL	= "tell";
  private static final String   PLAN_DATA = "PLAN-DATA";
  private static final String   SIM_DATA = "SIM-DATA";
  private static final String   DO_PLAN = "PLAN";
  private static final String   DO_SIM = "SIMULATE";


  // Names MPA is looking for
  private static final String TASK_NAME			= "AS-A";
  private static final String ADVICE_LABEL_1		= "advice1";
  private static final String ADVICE_LABEL_2		= "advice2";
  private static final String ADVICE_LABEL_3		= "advice3";
  private static final String PLAN_NAME_1		= "plan-advice1";
  private static final String PLAN_NAME_2		= "plan-advice2";
  private static final String PLAN_NAME_3		= "plan-advice3";
  // Not using these advice strings now, but we'll keep them around
  private static final String ADVICE_1			= "(" +
							  " :USE-TOMCATS-OVER-CARRIERS" +
							  " :USE-EAGLES-OVER-AIRFIELDS" +
							  " :INFLICT-BLINDNESS" +
							  " :DENY-AIR-PICTURE" +
							  " :INGRESS-AT-YUMA" +
							  " :KILL-SAMS" +
							  " :MASS-VS-SAM-LAUNCHERS" +
							  " :BE-NONPREEMPTIVE-ON-DEFENSE" +
							  ")";
  private static final String ADVICE_2			= "(" +
							  " :USE-TOMCATS-OVER-CARRIERS" +
							  " :USE-EAGLES-OVER-AIRFIELDS" +
							  " :INFLICT-BLINDNESS" +
							  " :DENY-AIR-PICTURE" +
							  " :INGRESS-AT-YUMA" +
							  " :KILL-SAMS" +
							  " :MASS-VS-SAM-LAUNCHERS" +
							  " :PREEMPT-ANTISHIP-THREATS" +
							  " :PREEMPT-BY-ATTACKING-MUNITIONS" +
							  " :PREEMPT-RUNWAYS" +
							  " :MASS-VS-AIRBASES" +
							  ")";
  private static final String ADVICE_3			= "(" +
							  " :IGNORE-ALL-SIDE-EFFECTS" +
							  " :USE-TOMCATS-OVER-CARRIERS" +
							  " :USE-EAGLES-OVER-AIRFIELDS" +
							  " :INFLICT-BLINDNESS" +
							  " :DENY-AIR-PICTURE" +
							  " :INGRESS-AT-YUMA" +
							  " :INGRESS-2-AT-SAN-FRANCISCO" +
							  " :KILL-SAMS" +
							  " :MASS-VS-SAM-LAUNCHERS" +
							  " :PREEMPT-ANTISHIP-THREATS" +
							  " :PREEMPT-BY-ATTACKING-MUNITIONS" +
							  " :PREEMPT-RUNWAYS" +
							  " :MASS-VS-AIRBASES" +
							  " :INGRESS-2-AT-SAN-FRANCISCO" +
							  " :BREACH-AT-TWO-PLACES" +
							  " :USE-BARRIER-CAPS" +
							  ")";

  // Names APAT is looking for
  private static final String APAT_PLAN_NAME		= "Granola Crunch";
  private static final String APAT_COA_NAME_1		= "COA1 Defensive";
  private static final String APAT_COA_NAME_2		= "COA2 PreEmptive";
  private static final String APAT_COA_NAME_3		= "COA3 DualFront";
  private static final String APAT_OBJECTIVE_NAME	= "Achieve Air Superiority";

  ////////////////////////////////////////////////////////////////////////
  // Class Variables
  ////////////////////////////////////////////////////////////////////////
  private static KqmlAgent agent = null;

	// TWP
	private static Vector mpaPlans;
	private static Vector simPlans;
	private static Vector mpaInWork;
	private static Vector simInWork;
	private static Vector taskVector;


  ////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////
  /**
   * Returns a MPA_AgentImpl and handles agent creation, registration,
   * and message i/o for APAT.
   */ 
  public MPA_AgentImpl() {
    
    if(DEBUG) System.out.println("Entering MPA_AgentImpl constructor");
	mpaPlans = new Vector();
	simPlans = new Vector();
	mpaInWork = new Vector();
	simInWork = new Vector();
	taskVector = new Vector();
    
    // Create the agent and turn on logging if requested
    try {
      if(DEBUG) System.out.print("Creating agent named: " + AGENT_APAT + " ... ");
      agent = new KqmlAgent(AGENT_APAT);
      if(DEBUG) System.out.println("done");

      if(LOGGING) {
	agent.enableLogging();
	Log.setLevel(LOGGING_LEVEL);
      }

      // Register the handlers for the KQML Performatives the agent understands
      //
      if(DEBUG) System.out.print("Registering handlers ... ");
      agent.registerHandler(PERFORMATIVE_EVALUATE,  (Handler)(this.new Performative_Evaluate()));
      agent.registerHandler(PERFORMATIVE_TELL,	    (Handler)(this.new Performative_Tell()));
      if(DEBUG) System.out.println("done");

    }
    catch (kqml.KqmlAgentRegistrationFailure e) { printException(e); }
    catch (kqml.KqmlAgentInitFailure e)         { printException(e); }

    if(DEBUG) System.out.println("Exiting MPA_AgentImpl constructor");
  }

  ////////////////////////////////////////////////////////////////////////
  // Class Methods
  ////////////////////////////////////////////////////////////////////////
  /**
   * This class method reads the SIPE-generated plan for  the
   * given plan->coa->objective using the given advice
   *
   *@param apatPlanName		the plan name from APAT
   *@param apatCoaName		the COA name from APAT
   *@param apatObjectiveName	the objective name from APAT
   */
  public static void readPlan(String apatPlanName,
			      String apatCoaName,
			      String apatObjectiveName) {
    
    if(DEBUG) System.out.println("Reading plan from Plan Server for Plan: \"" + apatPlanName  +
				 "\"  COA: \"" + apatCoaName +
				 "\"  Objective: \"" + apatObjectiveName);

   // Task and plan name SHOULD be retrieved from plan table
    String taskName = TASK_NAME;
    String planName = PLAN_NAME_1;
    readInPlanFromMPA(taskName, planName);

  }

  /**
   * This class method invokes the ACS Simulator to simulate the
   * previously chosen plans.
   *
   */
  public static void invokeSimulator()
	{
    
	// quick check to see if we're just wasting time
	if ( simPlans.size() == 0 )
	{
		System.out.println("No plans chosen to simulate");
		return;
	}

	String taskName = new String();
	String planName = new String();
	// ACS doesn't do multiple sim plans, so just get the guy off
	// the top
	String simData = new String();
	for (int x = 0; x < 1; x++)
	{
		simData = (String)simPlans.elementAt(x);
		simInWork.addElement( (String)simData );
	}

	// clear out the list of sim plans
	simPlans = new Vector();

	// HEY - this counts on the task being the first item in the
	// data, and the plan being the second.  If that changes so
	// should this.
	Vector temp = new Vector();
	StringTokenizer line = new StringTokenizer(simData,",");
	while ( line.hasMoreTokens() )
	{
		temp.addElement( line.nextToken() );
	}
	taskName = (String)temp.elementAt(0);
	planName = (String)temp.elementAt(1);
	// KLUDGE - Work around for the name differences between 
	//  APAT and the plan-server
	if (taskName.equals(APAT_PLAN_NAME))
		taskName = TASK_NAME;
	if ( planName.equals(APAT_COA_NAME_1) )
		planName = PLAN_NAME_1;
	else if ( planName.equals(APAT_COA_NAME_2) )
		planName = PLAN_NAME_2;
	else if ( planName.equals(APAT_COA_NAME_3) )
		planName = PLAN_NAME_3;


	// make sure the requested plan and coas is known by the plan-server.
	// if it isn't then send a warning message.
	System.out.println("Names should be translated now");
	System.out.println("SIM DATA - " + taskName + " " + planName);
	boolean validData = validateSimData(taskName, planName);

	if ( validData )
	{
    	sendMessageAndAwaitReply(AGENT_ACS, PERFORMATIVE_EVALUATE,
			 "(:SIMULATE " +
			 " :TASK " + quotedString(taskName) +
			 " :PLAN " + quotedString(planName) +
			 ")" 
			);
	}
	else
	{
		System.out.println("MPA_AGENT : Requested Plan and COA not found in" +
							"plan-server knowledge base");
	}
  }


	// It should be noted that this will invoke plan generation for 
	// multiple plans (apat->coas) for a single task (apat->plan).
	// We will not do multiple plan generation for multiple tasks.
	// You got that.
  /**
   * This class method invokes the MPA planner to generate a plan for the
   * plans that have been selected.  
   *
   */
	public static void invokePlanner()
	{

    if(DEBUG) System.out.println("Invoking Planner");

	// quick check to see if there are any plans
	if ( mpaPlans.size() == 0 )
	{
		System.out.println("No plans to invoke");
		return;
	}

	// now is the time when we go through the vector containing the
	// plans to generate.  All the plans should have the same task parent.
	// The first task found will be the one to compare against.  The
	// planning information will then be formatted so that the plan-server
	// can use it.
	String temp = (String)mpaPlans.elementAt(0);
	StringTokenizer line = new StringTokenizer(temp,",");
	String taskName = line.nextToken();
	String multiPlan = new String();
	String onePlan = new String();

	for ( int x = 0; x < mpaPlans.size(); x++)
	{
		String aPlan = (String) mpaPlans.elementAt(x);
		line = new StringTokenizer(aPlan, ",");
		String aTask = line.nextToken();
		if ( taskName.equals(aTask) )
		{
			mpaInWork.addElement(aPlan);
			String planName = line.nextToken();
			if ( planName.equals(APAT_COA_NAME_1) )
				planName = ADVICE_LABEL_1;
			else if ( planName.equals(APAT_COA_NAME_2) )
				planName = ADVICE_LABEL_2;
			else if ( planName.equals(APAT_COA_NAME_3) )
				planName = ADVICE_LABEL_3;
			String obj = line.nextToken();
			String advice = line.nextToken();
			onePlan = makePlanString(planName, advice);
			if ( x != (mpaPlans.size()-1) )
				multiPlan += (onePlan + " ");
			else 
				multiPlan += onePlan;
		}
		else
		{
			System.out.println("Found non-similar tasks in list");
		}
	}
	// Clean out the plans that were just submitted
	mpaPlans = new Vector();
	taskName = TASK_NAME;

	String planningmsg = "(:MULTIPLE-SOLVE :TASK " +
							quotedString(taskName) +
							" :ADVICE-CONTEXTS (" +
							multiPlan + "))";

    sendMessageAndAwaitReply(AGENT_PLAN_CELL_MGR, 
				PERFORMATIVE_EVALUATE, planningmsg);

  }

	/**
	 * Form a plan submission string for use with a multi-solve
	 *
	 *@param plan  The name of the plan
	 *@param advice The advice to use
	 */
	public static String makePlanString(String plan, String advice)
	{
		String theAdvice = "( " + advice + ")";
		String thePlan = quotedString(plan);
		String answer = "(" + thePlan + " " + theAdvice + ")";
		return answer;
	}

  /**
   * This class method updates the plan table once the MPA Planner
   * has completed its planning
   *
   *@param mpaTaskName		the task name from MPA
   *@param adviceLabel		the planning advice label set by APAT
   *@param apatPlanName		the name that MPA knows the plan by
   */
  public static void registerPlan(String mpaTaskName,
				  String adviceLabel,
				  String mpaPlanName) {

    // Find the plan entry for this advice label and register the plan name
    // Also, notify the user that the plan has been completed
  }


  ////////////////////////////////////////////////////////////////////////
  // Common KQML messages sent to MPA
  ////////////////////////////////////////////////////////////////////////
  /**
   * Sends a ping message to the specified agent
   *
   *@param destAgent	the agent to send ping to
   */
  public static void ping(String destAgent) {
    //sendMessage(destAgent, PERFORMATIVE_EVALUATE, "(:ping)");
    sendMessageAndAwaitReply(destAgent, PERFORMATIVE_EVALUATE, "(:ping)");
  }

  /**
   * Sends following KQML message to PlanServer to request plan
   *
   *@param taskName	the name of the task
   *@param planName	the name of the plan
   *@param viewName	the view for the plan
   */
  public static String queryPlan(String taskName, 
				 String planName, 
				 String viewName) {
    return(sendMessageAndAwaitReply(AGENT_PLAN_SERVER,
				    PERFORMATIVE_ASK_ALL,
				    "(:QUERY-PLAN" +
				    " :TASK " + quotedString(taskName) +
				    " :PLAN " + quotedString(planName) +
				    " :ACTION-NETWORK :ALL" +
				    " :VIEW " + viewName +
				    ")"
				    ));
  }

	/////////////////////////////////////////////////////////////
	// TWP methods
	/////////////////////////////////////////////////////////////
  /**
   * Requests all task names from the Plan Server and place them
   * in the task vector
   *
   */
  //private static void updateTaskList()
  public static void updateTaskList()
  {
	if(DEBUG) System.out.println("Updating list of tasks");
	// send a query to the plan-server for a list of tasks
    String tasks = sendMessageAndAwaitReply(AGENT_PLAN_SERVER,
                                    PERFORMATIVE_ASK_ALL,
                                    "(:QUERY-PLAN" +
                                    " :TASK :ALL" +
                                    " :VIEW :ASCII)");
	// parse the return string to get the tasks
	String one = My_Util.stripHeaderFromString(tasks);
	String two = My_Util.stripEndParens(one);
	taskVector = new Vector();
	StringTokenizer line = new StringTokenizer(two);
	while ( line.hasMoreTokens() )
	{
		taskVector.addElement( line.nextToken() );
	}
  }

	/**
	 * Requests all plans for a given task
	 *
	 *@param task  The task associated with the plans
	 */
	//private static Vector getTaskPlans(String task)
	public static Vector getTaskPlans(String task)
	{
		if(DEBUG) System.out.println("Getting plans for task" + task);
		Vector planVector = new Vector();
		if ( doesTaskExist(task) )
		{
			if(DEBUG) System.out.println("Valid task found");

			// send query to plan-server for list of plans
    		String plans = sendMessageAndAwaitReply(AGENT_PLAN_SERVER,
                                    PERFORMATIVE_ASK_ALL,
                                    "(:QUERY-PLAN" +
                                    " :TASK " +
									task +
									" :PLAN :ALL" +
                                    " :VIEW :ASCII)");
			// I'm not sure how to use the lisp utilities
			// so I'll massage the string another way
			String one = My_Util.stripHeaderFromString(plans);
			String two = My_Util.stripEndParens(one);
			StringTokenizer line = new StringTokenizer(two);
			while ( line.hasMoreTokens() )
			{
				planVector.addElement( line.nextToken() );
			}
		}
		return planVector;
	}

    /**
     * Determine if the task and plan exist in the plan server
     *
     *@param task  The task containing the plan
     *@param plan  The plan to be used to invoke the simulation
     */
    //private static boolean validateSimData(String task, String plan)
    public static boolean validateSimData(String task, String plan)
    {
        if ( doesPlanExist( quotedString(task), 
							quotedString(plan) ) )
        {
            return true;
        }
        else
        {
            if(DEBUG) System.out.println("Plan " + plan + " not known by plan-server");
            return false;
        }
    }
 
    /**
     * Determine if the task is known to the plan-server
     *
     *@param task  The name of the task to be found
     */
    //private static boolean doesTaskExist(String task)
    public static boolean doesTaskExist(String task)
    {
        updateTaskList();
		String atask = task;
        for (int x = 0; x < taskVector.size(); x++)
        {
            if (task.equals( (String)taskVector.elementAt(x) ) )
            {
                return true;
            }
        }
        return false;
    }
 
    /**
     * Determine if the plan exists for the given task
     *
     *@param task  The task name in the plan-server knowledge base
     *@param plan  The plan to be verified in the plan-server knowledge base
     */
    //private static boolean doesPlanExist(String task, String plan)
    public static boolean doesPlanExist(String task, String plan)
    {
        Vector planList = getTaskPlans(task);
        if ( planList.size() == 0 )
        {
            if(DEBUG) System.out.println("No plans found for task : " + task);
            return false;
        }
        for (int x = 0; x < planList.size(); x++)
        {
            if ( plan.equals( (String)planList.elementAt(x) ))
            {
                return true;
            }
        }
        return false;
    }

    /**
     * Return a vector of tasks contained in the plan-server
     *
     */
    public static Vector getTaskList()
    {
        updateTaskList();
        return taskVector;
    }

	/**
	 * Add the planning information to a vector that will later
	 * be used to start the planning process
	 *
	 *@param plan  The plan name ( mpa->task)
	 *@param coa   The coa name  ( mpa->plan)
	 *@param objective The objective that the plan wedge is for
	 *@param advice  The advice to create the plan
	 */
	public static void addMPAPlan(String planData)
	{
		mpaPlans.addElement(planData);
	}

	/**
	 * Add the simulation information to a vector that will be used
	 * to kick off the simulation process
	 *
	 *@param plan  The plan name (mpa->task)
	 *@param coa  The coa name (mpa->plan)
	 *@param objective The objective name - not really important now
	 */
	public static void addSimPlan(String simData)
	{
		simPlans.addElement(simData);
	}

	/**
	 * Let an instantiating application know if the agent is
	 * installed.
	 *
	 */
	public static boolean isInstalled()
	{
		return true;
	}


	/////////////////////////////////////////////////////////////
	// end TWP methods
	/////////////////////////////////////////////////////////////

  /**
   * Sends KQML message to the specified agent with the 
   * specified performative and contents
   *
   *@param destAgent	the agent to send msg to
   *@param performative	the message's performative
   *@param contents	the message's contents
   */
  public static void sendMessage(String destAgent, 
				 String performative, 
				 String contents) {
    String replyContents = null;

    try {
      Message msg = new Message (performative, contents, destAgent, agent);
      if(DEBUG) System.out.println("Sending message:\n  " + msg.toString());
      msg.send();
    }
    catch(kqml.KqmlException e)	    {printException(e);}
  }

 /**
   * Sends KQML message to the specified agent with the 
   * specified performative and contents
   *
   *@param destAgent	the agent to send msg to
   *@param performative	the message's performative
   *@param contents	the message's contents
   */
  public static String sendMessageAndAwaitReply(String destAgent, 
						String performative, 
						String contents) {
    String replyContents = null;

    try {
      Message msg = new Message (performative, contents, destAgent, agent);
      if(DEBUG) System.out.println("Sending message and awaiting reply:\n  " + msg.toString());
      Message reply = msg.sendAndAwaitReply();
		
      if(DEBUG) System.out.println("Received reply:\n  " +reply.toString());
      replyContents = reply.getContents();
		if(DEBUG) System.out.println("*****************************\n");
		if(DEBUG) System.out.println("Contents : <" + replyContents + ">");
		if(DEBUG) System.out.println("\n*****************************");
		return ( replyContents );
    }
    catch(kqml.KqmlException e)	    {printException(e);}
    return(replyContents);
  }


  ////////////////////////////////////////////////////////////////////////
  //
  // Once the plan is ready, 
  // 1) query the plan server for the files containing the 
  //    pre- and post-ctem plans
  // 2) invoke the Reader process on the files
  //
  ////////////////////////////////////////////////////////////////////////
  public static void readInPlanFromMPA(String taskName, String planName) {
    String replyContent = null;
    String preCTEM_Filename = null;
    String postCTEM_Filename = null;
    String subplans = null;
    String subplanName = null;

    // ASSUMING THAT ONLY ONE SUBPLAN FOR POST-CTEM PLAN
    replyContent = queryPlan(taskName, planName, ":subplans");
    if(Lisp.List.listp(replyContent) && (Lisp.List.nth(1, replyContent).equals(":QUERY-PLAN"))) 
	{
      subplans = Lisp.List.nth(2, replyContent);
      if(Lisp.List.listp(subplans) && Lisp.List.listLength(subplans) >= 1) 
	  {
		subplanName = unquotedString(Lisp.List.nth(1, subplans));
		if(DEBUG) System.out.println("Post-CTEM subplan for task \"" + 
						taskName + "\" and plan \"" + planName + 
				        "\" is \"" + subplanName +"\"");
      }
    }
    if(subplanName == null) 
		System.err.println("Expected Post-CTEM subplan for task \"" + 
							taskName + "\" and plan \"" + planName + 
					      "\"\n\tReceived \"" + replyContent +"\"");
	
    preCTEM_Filename = getPlanFilename(taskName, planName);
    if(subplanName != null) 
		postCTEM_Filename = getPlanFilename(taskName, subplanName);

	// Translate the names from plan-server to APAT
	String thePlanName = APAT_PLAN_NAME;
	String theCoaName = null;
	if ( planName.equals(PLAN_NAME_1) )
		theCoaName = APAT_COA_NAME_1;
	if ( planName.equals(PLAN_NAME_1) )
		theCoaName = APAT_COA_NAME_2;
	if ( planName.equals(PLAN_NAME_1) )
		theCoaName = APAT_COA_NAME_3;
	String theObjName = APAT_OBJECTIVE_NAME;

	// Using the task and plan name (apat->plan and coa)
	// determine which of the guys in the mpaInWork vector this is.
	// Fill in the info that is needed to invoke the reader with this
	// data and delete the member from the in-work vector.

	int planIndex = 0;
	boolean planFound = false;

	for (int x = 0; x < mpaInWork.size();x++)
	{
		String data = (String) mpaInWork.elementAt(x);
		StringTokenizer line = new StringTokenizer(data);
		String aplan = line.nextToken();
		String acoa = line.nextToken();
		String aobj = line.nextToken();
		if (theCoaName.equals(acoa))
		{
			planIndex = x;
			planFound = true;
			break;
		}
	}

	if ( planFound )
	{
		try { mpaInWork.removeElementAt(planIndex);}
		catch (ArrayIndexOutOfBoundsException e)
		{
			printException(e);
		}
	}

    if(DEBUG) System.out.println("Executing Reader on " + 
					preCTEM_Filename + " & " + postCTEM_Filename +
				 " for Plan:" + thePlanName +
				 " COA:" + theCoaName +
				 " OBJECTIVE:" + theObjName);

	Reader read = new Reader(preCTEM_Filename, postCTEM_Filename);
	read.processPlanData();
	read.createPlan(thePlanName, theCoaName, theObjName);
	
  }
  
  public static String getPlanFilename(String taskName, String planName) {
    String replyContent = null;
    String theFilename = null;

    replyContent = queryPlan(taskName, planName, ":ascii-filename");
    if(Lisp.List.listp(replyContent) 
        && (Lisp.List.nth(1, replyContent).equals(":QUERY-PLAN"))) 
	{
      theFilename = unquotedString(Lisp.List.nth(2, replyContent));
      if(DEBUG) System.out.println("Received ascii filename for task \"" + 
					taskName + "\" and plan \"" + planName + 
				   "\" is \"" + theFilename + "\"");
    }
    else 
	{
      System.err.println("Expected ascii filename for task \"" + taskName +
						 "\" and plan \"" + planName + 
						 "\"\n\tReceived \"" + replyContent +"\"");
    }
    return(theFilename);
  }
   

  public static boolean isError(Message aMsg) {
    return(false);
  }


  ////////////////////////////////////////////////////////////////////////
  // Incoming KQML Message Handlers
  ////////////////////////////////////////////////////////////////////////
  class Performative_Generic implements Handler {
    public Reply handler(Message msg) {
		if(DEBUG) System.out.println("Generic : Received message:\n  " 
										+ msg.toString() );
      return (Reply.reply("Received performative <" + msg.getPerformative() +
			  "> with contents <" +msg.getContents() +">"));
    }
  }

	class Performative_Tell extends Performative_Generic implements Handler 
	{
		public Reply handler(Message msg) 
		{
			if(DEBUG) System.out.println("Tell : Received message:\n  " + 
											msg.toString() );
			// what about the simulator?
			// May have to look at all the in work items (plans and sime)
			// and decide which one this is
			if(msg.getContents().regionMatches(true, 1, ":SOLUTION", 0, 9)) 
			{
				String data = msg.getContents();
				StringTokenizer line = new StringTokenizer(data);
				Vector items = new Vector();
				while ( line.hasMoreTokens() )
				{
					items.addElement( line.nextToken() );
				}
				// KLUDGE - this ain't so pretty
				String taskName = unquotedString((String)items.elementAt(2));
				String theLabel = unquotedString((String)items.elementAt(4));
				String thePlan = unquotedString((String)items.elementAt(6));
				readInPlanFromMPA(taskName, thePlan);
			}
			else
			{
				System.out.println("Unknown Tell Message received");
			}
			return(null);
		}
	}

	class Performative_Evaluate extends Performative_Generic implements Handler 
	{
		public Reply handler(Message msg) 
		{
			if(DEBUG) System.out.println("Evaluate : Received message:\n  " + 
											msg.toString() );
			String contents = msg.getContents();
			String msgType = getMsgType( contents );
			if ( msgType.equals(PLAN_DATA) )
			{
				if(DEBUG) System.out.println("Plan Data received");
				String temp = My_Util.stripHeaderFromString(contents);
				String data = My_Util.stripEndParens(temp);
				addMPAPlan(data);
				return Reply.reply("(:plan-data :ok)");
			}
			else if ( msgType.equals(SIM_DATA) )
			{
				if(DEBUG) System.out.println("Simulator Data received");
				String temp = My_Util.stripHeaderFromString(contents);
				String data = My_Util.stripEndParens(temp);
				addSimPlan(data);
				return Reply.reply("(:sim-data :ok)");
			}
			else if ( msgType.equals(DO_PLAN) )
			{
				if(DEBUG) System.out.println("Planning invocation received");
				invokePlanner();
				return Reply.reply("(:plan :ok)");
			}
			else if ( msgType.equals(DO_SIM) )
			{
				if(DEBUG) System.out.println("Simulator invocation received");
				invokeSimulator();
				return Reply.reply("(:simulate :ok)");
			}
			else if(msg.getContents().equals("(:ping)")) 
			{
				if(DEBUG) System.out.println("Received :ping message");
				return Reply.reply("(:ping :ok)");
			}
			else 
			{
				System.err.println("Unknown message:\n  " + msg.toString());
				return Reply.reply("Unknown content: <" + msg.getContents() +">");
			}
		}
	}

  ////////////////////////////////////////////////////////////////////////
  // Utility Methods
  ////////////////////////////////////////////////////////////////////////
  private static void printException(Exception e) { 
    System.out.println(e.getMessage()); 
  }

  private static String quotedString(String aString) { 
    return("\"" + aString + "\""); 
  }

  private static String unquotedString(String aString) {
    String tmpString = aString.trim();
    if(tmpString.charAt(0) == '"' 
        && tmpString.charAt(tmpString.length()-1) == '"')
      return(tmpString.substring(1,tmpString.length()-1));
    else return(null);
  }

	public static String getMsgType(String contents)
	{
		return ( My_Util.getKeyWord(contents) );
	}


  ////////////////////////////////////////////////////////////////////////
  // Main definition
  ////////////////////////////////////////////////////////////////////////
  public static void main(String args[]) 
  {

    if(DEBUG) System.out.println("Environment: " + System.getProperties());

    MPA_AgentImpl myComm = new MPA_AgentImpl();

    while (true) 
	{
      try
	  { Thread.sleep(1000); } 
      catch(InterruptedException e) 
	  { 
		printException(e);
	    System.out.println("Shutting down agent");
	    agent.shutdown();
	   }
    }
  }

  ////////////////////////////////////////////////////////////////////////
  // End of class definition
  ////////////////////////////////////////////////////////////////////////
}

