import kqml.Handler;
import kqml.KqmlAgent;
import kqml.Message;
import kqml.Reply;

public class TestServer 
{

	private static final String ASK_ONE     = "ask-one";
	private static final String EVALUATE    = "evaluate";
	private static final String DIE         = "die";
	private static final String TELL		= "tell";
	private static final long   SLEEP_TIME  = 1000;
	private static KqmlAgent    agent = null;

	public static void main(String args[]) 
	{
		//
		// Command line arguments:
		//	Arg 1: Agent Name
		//
		if (args.length < 1) 
			System.out.println("Usage: java TestServer <agentName>");
		else
			doServer(args[0]);
	}	

	public static void doServer(String serverName) 
	{
		try 
		{
			TestServer server = new TestServer();
			agent = new KqmlAgent(serverName);

			agent.registerHandler(ASK_ONE, (Handler)(server.new AskOne()));
			agent.registerHandler(EVALUATE, (Handler)(server.new Evaluate()));
			agent.registerHandler(DIE,     (Handler)(server.new Die()));
			agent.registerHandler(TELL,     (Handler)(server.new Tell()));

			//
			// Sleep in a loop forever.  If a "die" is received, the
			// die handler will shutdown and exit.
			//
			while (true) 
			{
				try { Thread.sleep(SLEEP_TIME); }
				catch(InterruptedException e) { printException(e); }
			}
	
		}
		catch (kqml.KqmlAgentRegistrationFailure e) { printException(e); }
		catch (kqml.KqmlAgentInitFailure e)         { printException(e); }
	}

	private static void printException(Exception e) 
	{
		System.out.println(e.getMessage());
	}
    

	class TestHandler implements Handler 
	{
		public Reply handler(Message msg) 
		{
			System.out.println("Received message:\n" + msg.toString());
			return (Reply.reply("Received performative <" + msg.getPerformative() +
				"> with contents <" +msg.getContents() +">"));
		}
	}

	class AskOne extends TestHandler implements Handler {}
	
	class Evaluate extends TestHandler implements Handler 
	{
		public Reply handler(Message msg) 
		{
			System.out.println("Received message:\n" + msg.toString());
			if( msg.getContents().equals("(:ping)") ) 
			{
				System.out.println("Received :ping message");
				return Reply.reply("(:ping :ok)");
			}
			else 
			{
				System.out.println("Unknown content");
				return Reply.reply("Unknown content: <" + msg.getContents() +">");
			}
		}
	}
    
	class Die extends TestHandler implements Handler 
	{
		public Reply handler(Message msg) 
		{
			super.handler(msg);
			System.out.println("Shutting down ...");
			agent.shutdown();	// must be a "nested" class to have "agent" // in scope
			System.exit(0);
			return (null);		// needed for the compiler
		}
	}

	class Tell extends TestHandler implements Handler 
	{
		public Reply handler(Message msg) 
		{
			System.out.println("Received message:\n" + msg.toString());
			if( msg.getContents().equals("(:ping)") ) 
			{
				System.out.println("Received :ping message");
				return Reply.reply("(:ping :ok)");
			}
			else 
			{
				System.out.println("Unknown content");
				System.out.println("Shutting down Server");
				agent.shutdown();
				return Reply.reply("Unknown content: <" + msg.getContents() +">");
			}
		}
	}
}
