/* File: ACP3_Watcher_Only.java
 * Contains: The main program for bin/acp3-watcher-only
 * Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Created: January 1998
 * Updated: Tue Jan 27 20:52:57 1998 by Jeff Dalton
 * Copyright: (c) 1998, AIAI, University of Edinburgh
 */


package oplan.acp3;

import oplan.util.*;
import kqml.*;


/** The ACP3 application Kqml-watcher-only main program, for testing . */

public class ACP3_Watcher_Only extends ACP3 {

    static ProcessViewer viewer;
    static KqmlWatcher watcher;
    static Thread commandLoop;

    public static void main(String[] argv) {

	printGreeting("ACP3 KQML-Watcher only");

	System.out.println("Type \"exit\" to exit.\n");

	// Create an object to represent the ACP3 application
	agent = new ACP3_Watcher_Only();

	// Create a trivial viewer
	viewer = new TrivialViewer();

	// Start a trivial command-loop thread so the user can
	// tell us to exit.
	commandLoop = new TrivialCommandLoop();
	commandLoop.start();

	// Create a Kqml watcher and let it take over.
	watcher = new KqmlWatcher(viewer);
	watcher.init();
	watcher.watcherMainLoop();
	
    }

    protected void shutdown() {
	if (watcher != null)
	    watcher.shutdown();
	super.shutdown();
    }

}


class TrivialViewer implements ProcessViewer {

    public void startAction(String action, int coa) {
	Debug.noteln("\n>>> Start of action: " + action + " " + coa + "\n");
    }

}


class TrivialCommandLoop extends Thread {

    // The Kqml message-processing threads don't seem to run if
    // we use Util.readLine(System.in) here.  Util.askLine has a
    // loop with 1-second sleeps, which seems to work but shouldn't
    // be necessary.

    public void run() {
	for (;;) {
	    String command = Util.askLine("Command:");
	    if (command.equals("bye")
		|| command.equals("exit")
		|| command.equals("quit"))
		ACP3.exit(0);
	    else if (command.equals("ping"))
		sendPing();
	}
    }

    public static void sendPing() {
	String who = Util.askLine("Name of agent to ping:");
	try {
	    Message m
		= new Message("evaluate",
			      "(:ping)",
			      who, 
			      ACP3_Watcher_Only.watcher.kqmlAgent);
	    // m.setKeywordValue(":reply-with", ":ping");
	    // m.send();
	    // /*
	    Message reply
		= m.sendAndAwaitReply();
	    Debug.noteln("Reply:", reply);
	    // */
	}
	catch (KqmlException e) {
	    Debug.noteException(e);
	}
    }

}
