/*
** Restricted Rights Legend
**
** Use, duplication or disclosure by the government is subject to restrictions
** as set forth in subparagraph (c)(1)(ii) of the Rights in Technical Data and
** Computer Software clause at DFARS 252.227-7013.
**
** Copyright (c) 1997 Lockheed Martin C2 Integration Systems
**
**
** File:         MsgReceiver.java
**
** Description:  Contains the MsgReceiver class that is internal to the KRIL.
**
*/
package kqml;

/**
 * A class internal to the KRIL.
 *
 * This class is run as a thread and used to poll the connections
 * in the cache to determine if an input message is present on the
 * connection.
 * 
 * @author Peter Stevens
 * @version $Revision: 1.7 $
 * @see KqmlAgent#KqmlAgent(java.lang.String, int, java.lang.String)
 */

class MsgReceiver extends Thread
{
    private KqmlAgent agent = null;
    private static final long SLEEP_TIME = 50;

    // Constructor
    MsgReceiver(KqmlAgent agent)
    {
        super("MsgReceiver");
        this.agent = agent;
    }

    public void run()
    {
        setPriority(getPriority() + 2);
        //
        // Search through the connection cache for any input messages
        // - all of the connections in the cache are checked for a message
        //
        Log.log(Log.THREAD, "Starting Receiver for " + agent.getAgentName());

        while (true)
        {
            agent.scanCacheForMsgs();
            try
            {
                Thread.sleep(SLEEP_TIME);
            }
            catch(InterruptedException e)
            {
            }
        }
    }
}
