/*** File: Process.java
*** Contains: Class defining the main process element for CPE.
*** Author: Steve Polyak <Steve_Polyak@ed.ac.uk>
*** Created: June 1997
*** Updated: Wed Jan 21 19:15:49 1998 by Jeff Dalton
*/

package oplan.acp3;

import java.util.*;
import java.awt.*;
import oplan.util.*;
// import Node;
// import Edge;
// import ViewGraph;

class CPEProcess {
  //Link to the panel and node property dialog
  ViewGraph m_panel = null;
  
  //Elements kept as a vector of nodes, edges
  Vector m_nodes;
  Vector m_edges;
  
  //Constants
  final static int FINISH_NODE_SPACING = 15;
  final static String FINISH_STRING = "Finish";
  final static String START_STRING = "Start";

  //Graphics elements
  final Color m_edgeColor = Color.black;
  final Color m_linkColor = Color.lightGray;
  Image m_offscreen;
  Dimension m_offscreensize;
  Dimension m_dimension;
  Graphics m_offgraphics;
  Rectangle m_selectingRect;

  //Flags and Counters
  long m_nodeCounter = 1;
  boolean m_linkDrawing = false;
  boolean m_selectingBegin = false;
  boolean m_selectingEnd = false;
  Node m_begin = null;
  Node m_hitNode = null;
  boolean m_selecting = false;
  int m_linkToX = 0;
  int m_linkToY = 0;
  
  //Constructor saves a reference back to the panel.
  CPEProcess(ViewGraph panel) {
    m_panel = panel;
    m_nodes = new Vector();
    m_edges = new Vector();		
  }
  
  public void clear() {
    m_edges.removeAllElements();
    m_nodes.removeAllElements();  
  }
  
  public void newProcess() {
    m_dimension = m_panel.size();
    clear();
    addNode(new SpecialNode(FINISH_STRING,
			    ((int) (FINISH_NODE_SPACING-1)
			     *(m_dimension.width / FINISH_NODE_SPACING)), 
			    (int) m_dimension.height/2));
    addNode(new SpecialNode(START_STRING, 
			    ((int) m_dimension.width/FINISH_NODE_SPACING), 
			    (int) m_dimension.height/2));	
    addEdge(START_STRING, FINISH_STRING);
  }
  
  void moved() {
    
    m_dimension = m_panel.size();
    Node tmpNode = locateNode(FINISH_STRING);
    
    if (tmpNode != null) {
      tmpNode.m_x = (int) (FINISH_NODE_SPACING-1) *
	(m_dimension.width /FINISH_NODE_SPACING);
      tmpNode.m_rect.x = tmpNode.m_x;
      tmpNode.m_y = (int) m_dimension.height/2;
      tmpNode.m_rect.y = tmpNode.m_y;
    }
    tmpNode = locateNode(START_STRING);
    if (tmpNode != null) {
      tmpNode.m_x = (int) (m_dimension.width /FINISH_NODE_SPACING);
      tmpNode.m_rect.x = tmpNode.m_x;
      tmpNode.m_y = (int) m_dimension.height/2;
      tmpNode.m_rect.y = tmpNode.m_y;
    }
    
  }
  
  void addNode(Node node) {
    m_nodes.addElement(node);
  }
  
  void removeSelected() {
    Vector delNodes = new Vector();
    Node tempNode;
    for  (Enumeration e =  m_nodes.elements() ; 
	  e.hasMoreElements()  ;)  {	
      tempNode = (Node) e.nextElement();
      if (tempNode.m_selected) {
        delNodes.addElement(tempNode);
      }
    }
    for  (Enumeration  e2  =  delNodes.elements() ;
	  e2.hasMoreElements()  ;)  {	
      tempNode = (Node) e2.nextElement();
      removeNode(tempNode);
    }
    delNodes = null;
  }
  
  void removeNode(Node node) {
    Edge tempEdge;
    
    if ((node.m_key.equals("")) || 
        (node.m_key.equals(START_STRING)) ||
        (node.m_key.equals(FINISH_STRING))) {
      return;
    }
    
    //no need to search for node, just search 
    // for any edges.
    Vector delEdges = new Vector();
    for  (Enumeration  e1  =  m_edges.elements() ;
	  e1.hasMoreElements()  ;)  {	
      tempEdge = (Edge) e1.nextElement();
      if ((tempEdge.m_from.m_key.equals(node.m_key)) || 
	  (tempEdge.m_to.m_key.equals(node.m_key))) {
	delEdges.addElement(tempEdge);
      }
    }
    for  (Enumeration  e2  =  delEdges.elements() ; 
	  e2.hasMoreElements()  ;)  {	
      tempEdge = (Edge) e2.nextElement();
      m_edges.removeElement(tempEdge);
    }
    m_nodes.removeElement(node);
    delEdges = null;
  }
  
  void renameNode(String from, String to) 
  {
    Node temp;
    if ((from.equals("")) || 
        (to.equals("")) ||
        (from.equals(START_STRING)) ||
        (to.equals(START_STRING)) ||
        (from.equals(FINISH_STRING)) ||
        (to.equals(FINISH_STRING)) ||
        (locateNode(to) != null)) {
      return;
    }
    
    //For now, a simple linear search
    for  (Enumeration  e  =  m_nodes.elements() ;
	  e.hasMoreElements()  ;)  {	
      temp = (Node) e.nextElement();
      if (temp.m_key.equals(from)) {
        temp.m_lbl = to;
        break; 
      }
    }
  }
  
  Node locateNode(String key) {
    Node found = null;
    Node temp;
    
    //For now, a simple linear search
    for  (Enumeration  e  =  m_nodes.elements() ;
	  e.hasMoreElements()  ;)  {	
      temp = (Node) e.nextElement();
      if (temp.m_key.equals(key)) {
        found = temp;
        break; 
      }
    }	
    return found;
  }

  void addEdge(String from, String to) {
    Node nodeFrom, nodeTo;
    nodeFrom = locateNode(from);
    nodeTo = locateNode(to);
    
    m_edges.addElement(new Edge(nodeFrom, nodeTo));
  }
  
  void removeEdge(String from, String to) {
    Edge temp;
    //For now, a simple linear search
    for  (Enumeration  e  =  m_edges.elements() ;
	  e.hasMoreElements()  ;)  {	
      temp = (Edge) e.nextElement();
      if ((temp.m_from.m_key.equals(from)) &&
	  (temp.m_to.m_key.equals(to))) {
        m_edges.removeElement(temp);
        break; 
      }
    }
  }
  
  void update(Graphics g ) {
    
    Dimension d = m_panel.size();    
    
    if ((m_offscreen == null) || 
	(d.width != m_offscreensize.width) ||
	(d.height != m_offscreensize.height)) {
      m_offscreen = m_panel.createImage(d.width, d.height);
      m_offscreensize = d;
      m_offgraphics = m_offscreen.getGraphics();
      m_offgraphics.setFont(m_panel.getFont());
    }
    // moved();
    
    m_offgraphics.setColor(m_panel.getBackground());
    m_offgraphics.fillRect(0, 0, d.width, d.height);
    FontMetrics fm = m_offgraphics.getFontMetrics();
    
    // Draw each edge
    for  (Enumeration  e  =  m_edges.elements() ;
	  e.hasMoreElements()  ;)  {	
      ((Edge) e.nextElement()).draw(m_offgraphics, fm);
    }	
    
    // Draw each node.
    for  (Enumeration  e  =  m_nodes.elements() ;
	  e.hasMoreElements()  ;)  {	
      ((Node) e.nextElement()).draw(m_offgraphics, fm);
    }
    
    // Draw each edge arrow
    for  (Enumeration  e  =  m_edges.elements() ;
	  e.hasMoreElements()  ;)  {	
      ((Edge) e.nextElement()).drawArrow(m_offgraphics, fm);
    }	
    
    //Put on the 3d frame
    //m_offgraphics.setColor(m_panel.getBackground());
    //m_offgraphics.draw3DRect(0, 0, d.width - 1, d.height - 1, true);
    //m_offgraphics.draw3DRect(3, 3, d.width - 7, d.height - 7, false);
    
    //Link drawing and selection goes on top.
    if (m_linkDrawing) {
      m_offgraphics.setColor(m_linkColor);
      m_offgraphics.drawLine(m_hitNode.m_x, m_hitNode.m_y,
			     m_linkToX, m_linkToY);
    }

    if (m_selecting) {
      Rectangle box = getDrawableRect(m_selectingRect, d);
      m_offgraphics.setColor(m_linkColor);
      m_offgraphics.drawRect(box.x, box.y, 
			     box.width - 1, box.height - 1);
    }
    
    g.drawImage(m_offscreen, 0, 0, null);
  }

  public boolean mouseDown(Event evt, int x, int y) {
    Node temp;
    m_hitNode = null;
    
    //look for a hit node (in reverse for proper Z order)
    for(int intSize = m_nodes.size(); intSize > 0; intSize--) {
      temp = (Node) m_nodes.elementAt(intSize - 1);
      if (temp.m_rect.inside(x,y)) {
	m_hitNode = temp;
	break;	
      }
    }
    
    if (m_selectingBegin) {
      if (m_hitNode != null)
	handleBeginNodeSelecting(m_hitNode);
      else {
	m_selectingBegin = false;
	//update status label
	updateStatus("Ready...", Color.lightGray);
      }
    }
    else if (m_selectingEnd) {
      if (m_hitNode != null)
	handleEndNodeSelecting(m_hitNode);
      else {
	m_selectingEnd = false;
	m_begin.m_beginSelected = false;
	m_begin = null;
	//update status label
	updateStatus("Ready...", Color.lightGray);
      }
    }
    else if (evt.shiftDown() && m_hitNode != null) {
      //alt key was used on a node click, pop node dialog
      Debug.noteln("node editing disabled in ACP3");
      return true;
    }
    else if (evt.metaDown() && m_hitNode == null) {
      m_panel.repaint();
      return true;
    }
    else if(evt.metaDown() && m_hitNode != null) {
      // going into link draw mode
      m_linkDrawing = true;
      m_linkToX = x;
      m_linkToY = y;
    }
    else if(m_hitNode != null) {
      if ((m_hitNode.m_key == START_STRING) ||
	  (m_hitNode.m_key == FINISH_STRING)) {
	m_hitNode = null;
      }
      else {
	if(evt.controlDown() == false)
	  deselectAll();
	m_hitNode.m_selected = true;
      }
    }
    else {
      //going into selection mode
      if ((evt.metaDown() != true) && (m_selectingBegin != true) &&
	  (m_selectingEnd != true)) {
	deselectAll();
	m_selecting = true;
	m_selectingRect = new Rectangle(x, y, 0, 0);
      }
    }
    m_panel.repaint();
    return true;
  }
  
  public boolean mouseDrag(Event evt, int x, int y) {
    
    if (m_panel.inside(x,y)) {
      if (m_linkDrawing) {
	m_linkToX = x;
	m_linkToY = y;
      }
      else if (m_selecting) {
	m_selectingRect.resize(x - m_selectingRect.x,
			       y - m_selectingRect.y);
      }
      else {
	if (m_hitNode != null) {
	  if ((m_hitNode.m_lbl == START_STRING) ||
	      (m_hitNode.m_lbl == FINISH_STRING)) {
	    m_hitNode = null;
	  }
	  else {
	    if ((evt.shiftDown() == false) &&
		(m_selectingEnd == false) &&
		(m_selectingBegin == false)) {
	      m_hitNode.m_x = x;
	      m_hitNode.m_y = y;
	    }
	  }
	}
      }
    }
    m_panel.repaint();
    return true;
  }
  
  public boolean mouseMove(Event evt, int x, int y) {
    Node tempNode;
    
    if(m_linkDrawing || m_selecting) {
      m_linkDrawing = false;
      m_selecting = false;
    }
    if (m_selectingBegin || m_selectingEnd) {
      for  (Enumeration  e  =  m_nodes.elements() ;
	    e.hasMoreElements()  ;)  {	
	tempNode = (Node) e.nextElement();
	if (tempNode.m_rect.inside(x,y)) {
	  if (m_selectingBegin)
	    tempNode.m_beginFocus = true;
	  else
	    tempNode.m_endFocus = true;
	}
	else {
	  tempNode.m_beginFocus = false;
	  tempNode.m_endFocus = false;
	}
      }
    }
    return true;
  }
  
  public boolean mouseUp(Event evt, int x, int y) {

    if (m_panel.inside(x,y)) {
      if(m_linkDrawing) {
	m_linkDrawing = false;
	linkDrawUp(x,y);
      }
      else if (m_selecting) {
	m_selecting = false;
	selectingUp(x,y);
      }
      else {}
    }
    m_hitNode = null;
    m_panel.repaint();
    return true;
  }
	
  void linkDrawUp(int x, int y) {
    Node tempNode;
    Edge tempEdge;
    boolean duplicateEdge = false;
    
    //look for the node that was released on
    for(int intSize = m_nodes.size(); intSize > 0; intSize--) {
      tempNode = (Node) m_nodes.elementAt(intSize - 1);
      if (tempNode.m_rect.inside(x,y)) {
	//found node, make sure edge doesn't exist
	for  (Enumeration  e1  =  m_edges.elements() ;
	      e1.hasMoreElements()  ;)  {	
	  tempEdge = (Edge) e1.nextElement();
	  if (((tempEdge.m_from.m_key == m_hitNode.m_key) &&
	       (tempEdge.m_to.m_key == tempNode.m_key)) 
	      ||
	      ((tempEdge.m_from.m_key == tempNode.m_key) &&
	       (tempEdge.m_to.m_key == m_hitNode.m_key))) {
	    duplicateEdge = true;
	    break;
	  }
	}
	if (duplicateEdge == false) {
	  addEdge(m_hitNode.m_key, tempNode.m_key);
	}
	break;	
      }
    }
  }
  
  void selectingUp(int x, int y) {
    Node tempNode;
    Edge tempEdge;
    Rectangle box = getDrawableRect(m_selectingRect, m_dimension);
    
    //find nodes that intersect with the selection
    for  (Enumeration  e  =  m_nodes.elements() ;
	  e.hasMoreElements()  ;)  {	
      tempNode = (Node) e.nextElement();
      if (tempNode.m_rect.intersects(box)){
        if ((tempNode.m_key.equals(START_STRING) == false) &&
            (tempNode.m_key.equals(FINISH_STRING) == false)) {
          tempNode.m_selected = true;
        }
      }
    }
  }
  
  public void addDummy()
  {
    String newString = new String("Dummy" );
    newString = newString.concat(Long.toString(m_nodeCounter));
    addNode(new SpecialNode(newString, 
			    ((int)m_dimension.width/2), 
			    (int) m_dimension.height/2));
    m_nodeCounter++;
  }

  public void addActivity(){
    
    if (m_nodes.size() < 3) {
      String newString = new String("Act" );
      newString = newString.concat(Long.toString(m_nodeCounter));
      removeEdge(START_STRING, FINISH_STRING);
      addNode(new Node(newString, 
		       ((int)m_dimension.width/2), 
		       (int) m_dimension.height/2));
      addEdge(START_STRING, newString);
      addEdge(newString, FINISH_STRING);
      m_nodeCounter++;
    }
    else {
      //Need to go into a begin/end node selection
      if (m_selectingBegin != true && m_selectingEnd != true) {
        m_selectingBegin = true;
        deselectAll();
        //update status label
        updateStatus("Select begin", Color.yellow);
      }
    }
  }
  
  Rectangle getDrawableRect(Rectangle originalRect, 
			    Dimension drawingArea) {
    int x = originalRect.x;
    int y = originalRect.y;
    int width = originalRect.width;
    int height = originalRect.height;
    
    //Make sure rectangle width and height are positive.
    if (width < 0) {
      width = 0 - width;
      x = x - width + 1;
      if (x < 0) {
        width += x;
        x = 0;
      }
    }
    if (height < 0) {
      height = 0 - height;
      y = y - height + 1;
      if (y < 0) {
        height += y;
        y = 0;
      }
    }
    
    //The rectangle shouldn't extend past the drawing area.
    if ((x + width) > drawingArea.width) {
      width = drawingArea.width - x;
    }
    if ((y + height) > drawingArea.height) {
      height = drawingArea.height - y;
    }
    return new Rectangle(x, y, width, height);
  }
  
  void deselectAll() {
    for  (Enumeration  e  =  m_nodes.elements() ;
	  e.hasMoreElements()  ;)  {	
      ((Node) e.nextElement()).m_selected = false;
    }
  }
  
  void handleBeginNodeSelecting(Node node) {
    m_begin = node;
    m_begin.m_beginSelected = true;
    m_begin.m_beginFocus = false;
    m_selectingEnd = true;
    m_selectingBegin = false;
    //update status label
    updateStatus("Select end", Color.yellow);
  }
  
  void handleEndNodeSelecting(Node node) {
    int newx, newy;
    String newString = new String("Act" );
    
    newString = newString.concat(Long.toString(m_nodeCounter));
    //put the new node halfway between the two
    if (node.m_x >= m_begin.m_x)
      newx = m_begin.m_x + ((int)(node.m_x - m_begin.m_x)/2);
    else
      newx = node.m_x + ((int)(m_begin.m_x - node.m_x)/2);
    if (node.m_y >= m_begin.m_y)
      newy = m_begin.m_y + ((int)(node.m_y - m_begin.m_y)/2);
    else
      newy = node.m_y + ((int)(m_begin.m_y - node.m_y)/2);
    
    addNode(new Node(newString, newx, newy));
    addEdge(m_begin.m_key, newString);
    addEdge(newString, node.m_key);
    m_nodeCounter++;
    
    //if a link existed between the two action
    //before, remove it.
    removeEdge(m_begin.m_key, node.m_key);
    
    // clear focus of begin and end node
    node.m_endFocus = false;
    m_begin.m_beginSelected = false;
    // clear temp begin node holder
    m_begin = null;
    // turn off process-wide flag
    m_selectingEnd = false;
    // clear hitnode
    m_hitNode = null;
    //update status label
    updateStatus("Ready...", Color.lightGray);
  }

  //updates the status bar description
  public void updateStatus(String message, Color color) { 
    Debug.noteln
      ("CPEProcess.updateStatus(" + message + "," + color + ") called");
  }
  
}
