/* File: LispTest.java
 * Contains: Main program for testing Lisp classes.
 * Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Created: January 1998
 * Updated: Thu Jan 22 18:10:50 1998 by Jeff Dalton
 * Copyright: (c) 1998, AIAI, University of Edinburgh
 */

package oplan.test;

import java.util.*;
import oplan.util.*;
import oplan.util.lisp.*;


/** A class for testing */

class LispTest {

    public static void main(String[] argv) {

	Debug.noteln("An empty list:", Lisp.list());

	Debug.noteln("A list of 3 strings:", Lisp.list("a", "b", "c"));

	Debug.noteln("Length of () =", Lisp.list().length());

	Debug.noteln("Length of (a b) =", Lisp.list("a", "b").length());

	Debug.assert(Symbol.intern("apple") == Symbol.intern("apple"));

	Debug.noteln("A keyword:", Symbol.intern(":apple"));

	{
	    Symbol[] syms = {Symbol.intern("apple"),
			     Symbol.intern(":apple"),
			     Symbol.intern("?apple"),
			     Keyword.intern(":apple")};
	    for (int sy = 0; sy < syms.length; sy++) {
		Debug.noteln("Class of " + syms[sy] + " ==> " +
			     syms[sy].getClass());
	    }
	}

	// What if we try to instantiate List?  The compiler objects
	//   (as it should).
	// Debug.noteln("A List:", new List());

	Debug.noteln("Enumerations:");
	printEnumeration(Lisp.list().elements());
	printEnumeration(Lisp.list("a", "b", "c").elements());

	Debug.noteln("elementAt:");
	{
	    List l = Lisp.list("a", "b", "c");
	    Debug.noteln("0: " + l.elementAt(0));
	    Debug.noteln("1: " + l.elementAt(1));
	    Debug.noteln("2: " + l.elementAt(2));
	    Debug.noteln("3: " + l.elementAt(3));

	    Debug.noteln("1st elt of nil: " + Lisp.list().elementAt(1));
	}

	Debug.noteln("Append (a b c) (d e) ==>",
		     Lisp.list("a", "b", "c").append(Lisp.list("d", "e")));
	Debug.noteln("Append () (d e) ==>",
		     Lisp.list().append(Lisp.list("d", "e")));

	Debug.noteln("Equal (a b c) (a b c) ==> " +
		     Lisp.equal(Lisp.list("a", "b", "c"),
				Lisp.list("a", "b", "c")));
	Debug.noteln("Equal (a b c) (a c b) ==> " +
		     Lisp.equal(Lisp.list("a", "b", "c"),
				Lisp.list("a", "c", "b")));
	Debug.noteln("Equal (a b c) (a b) ==> " +
		     Lisp.equal(Lisp.list("a", "b", "c"),
				Lisp.list("a", "b")));
	Debug.noteln("Equal () () ==> " +
		     Lisp.equal(Lisp.list(), Lisp.list()));


	Debug.noteln("find () a ==> " + Lisp.list().find(Symbol.intern("a")));
	Debug.noteln("find (a b c) a ==> " +
		     stringToList("(a b c)").find(Symbol.intern("a")));
	Debug.noteln("find (a b c) b ==> " +
		     stringToList("(a b c)").find(Symbol.intern("b")));
	Debug.noteln("find (a b c) c ==> " +
		     stringToList("(a b c)").find(Symbol.intern("c")));
	Debug.noteln("find (a b c) d ==> " +
		     stringToList("(a b c)").find(Symbol.intern("d")));

	Debug.noteln("get (a b c) a ==>",
		     stringToList("(a b c)").get(Symbol.intern("a")));
	Debug.noteln("get (a b c) b ==>",
		     stringToList("(a b c)").get(Symbol.intern("b")));
	Debug.noteln("get (a b c) c ==>",
		     stringToList("(a b c)").get(Symbol.intern("c")));


	Debug.noteln("Reverse (a b c) ==>",
		     stringToList("(a b c)").reverse());

	Debug.noteln("Take (a b c) 0 ==>",
		     stringToList("(a b c)").take(0));
	Debug.noteln("Take (a b c) 1 ==>",
		     stringToList("(a b c)").take(1));
	Debug.noteln("Take (a b c) 2 ==>",
		     stringToList("(a b c)").take(2));
	Debug.noteln("Take (a b c) 3 ==>",
		     stringToList("(a b c)").take(3));
	Debug.noteln("Take (a b c) 4 ==>",
		     stringToList("(a b c)").take(4));

	Debug.noteln("Drop (a b c) 0 ==>",
		     stringToList("(a b c)").drop(0));
	Debug.noteln("Drop (a b c) 1 ==>",
		     stringToList("(a b c)").drop(1));
	Debug.noteln("Drop (a b c) 2 ==>",
		     stringToList("(a b c)").drop(2));
	Debug.noteln("Drop (a b c) 3 ==>",
		     stringToList("(a b c)").drop(3));
	Debug.noteln("Drop (a b c) 4 ==>",
		     stringToList("(a b c)").drop(4));

    }

    static List stringToList(String s) {
	return (List)Lisp.readFromString(s);
    }

    static void printEnumeration(Enumeration e) {
	Debug.note("[");
	for (int i = 0; e.hasMoreElements(); i++) {
	    Debug.note(i + ": " + e.nextElement() + "\n ");
	}
	Debug.noteln("]");
    }

}
