/* File: MatchTest.java
 * Contains: Main program for testing MatchTables
 * Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Created: January 1998
 * Updated: Mon Jan 19 23:04:52 1998 by Jeff Dalton
 * Copyright: (c) 1998, AIAI, University of Edinburgh
 */

package oplan.test;

import java.util.*;
import oplan.util.*;
import oplan.util.match.*;
import java.io.*;


/** A class for testing MatchTable pattern-matching. */

class MatchTest {

    public static void main(String[] argv) {

	StringMatchTable table = new StringMatchTable();

	table.addPattern("apple");
	table.addPattern("pie");

	/* Now loop, matching against what the user types. */

	String input;
	Object matchResult;

	System.out.println("Type \"bye\" to exit.\n");
	for (;;) {
	    input = ask("How about:");
	    if (input.equals("bye"))
		break;
	    matchResult = table.match(input);
	    if (matchResult == null)
		System.out.println("Nope.");
	    else
		System.out.println(matchResult);
	}
	System.out.println("Bye!\n");

    }

    static String ask(String prompt) {
	System.out.print(prompt + " ");
	System.out.flush();
	return Util.readLine(System.in);
    }

}


/* StringMatchTable's addPattern automatically adds an appropriate
   MatchCase, given a String pattern. */

class StringMatchTable extends MatchTable {

    void addPattern(String pattern) {
	addCase(new StringMatchCase(pattern));
    }

}


/* StringMatchCase provides incredibly simple matching using the
   String equals method. */

class StringMatchCase extends MatchCase {

    StringMatchCase(String pattern) {
	this.pattern = pattern;
    }

    public Object tryMatch(Object data) {
	if (pattern.equals(data))
	    return data;
	else
	    return null;
    }

    public Object selected(Object matchResult) {
	return "Cool: " + matchResult;
    }

}
