/* File: WaitImage.java
 * Contains: Blocking Image Loader, and defualt image creation
 * Author: Austin Tate <a.tate@ed.ac.uk>
 * Created: Thu Jan 15 16:39:25 1998
 * Updated: Thu Jan 22 17:20:28 1998 by Jeff Dalton
 * Copyright: (c) 1998, AIAI, University of Edinburgh
 */

package oplan.util;

import java.awt.*;
import java.awt.image.*;


public class WaitImage {

    public static int DEFAULT_IMAGE_SIZE=12;

    /** Create in memory a default image in a nominated component.
        Component must have a peer when this is called.
        Image is a square transparent (clear black) image with
        a gray cross on it */

    public static Image defaultImage(Component c) {
        Image img;
        MediaTracker mt = new MediaTracker(c);
        int sz = DEFAULT_IMAGE_SIZE;   //for convenience of short variable name
        int[] pixels = new int[sz*sz]; //initialised to 0 (clear black)
        for (int i = 0; i<sz; i++) {
            pixels[(i*sz)+i] = 0xFFCCCCCC; //initial FF for opaqueness
            pixels[((i+1)*sz)-i] = 0xFFCCCCCC;
        }
        Debug.note("not available. Create default...");
        MemoryImageSource imgs = new MemoryImageSource(sz,sz,pixels,0,sz);
        img = c.createImage(imgs);
        mt.addImage(img, 0);
        try { mt.waitForID(0); }
        catch (InterruptedException ex) {
            Debug.note("failed.");
            return null;
        };
        return img;
    }

    /** Method to load an image for a component,
        waiting for it to become available.
        If any error, substitute a small internally crafted image
        clear with a gray cross on it */

    public static Image waitGetImage (String fname, Component c) {
        Image img = c.getToolkit().getImage(fname);
        MediaTracker mt = new MediaTracker(c);
        mt.addImage(img, 0);
        Debug.note("Loading image " + fname + "...");
        try { mt.waitForID(0); }
	catch (InterruptedException ex) {} // ignore interrupts
        if (mt.isErrorID(0)) {
           img = defaultImage(c);
        }
        Debug.noteln("done.");
        return img;
    }

    /** waitCreateImage is used for secure image handling in ViewMatrix.
      It makes sure we WAIT for images to be created rather than
      relaying on the AWT thread to queue up graphics operations until
      the image is created in the peer. */

    public static Image waitCreateImage (int w, int h, Component c) {
	Image img = c.createImage(w,h);
	MediaTracker mt = new MediaTracker(c);
	mt.addImage(img, 0);
	// Debug.note("Creating image "+w+"x"+h+"...");
	try {mt.waitForID(0);
	} catch (InterruptedException ex) {} //ignore exceptions;
	if (mt.isErrorID(0)) {
	    img = defaultImage(c);
	}
	// Debug.noteln("done.");
	return img;
    }

}
